# suikyo.rb: Romaji-Hiragana conversion library for Ruby
# $Id: suikyo.rb,v 1.1 2003/06/02 17:36:24 komatsu Exp $
#
# Copyright (C) 2002 Hiroyuki Komatsu <komatsu@taiyaki.org>
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

$KCODE = 'e'
require 'kconv'
require 'suikyo/suikyo-config'

class File
  def File::join2 (*paths)
    dirs = paths[0..-2].map{|path|
      path ? path.split(File::Separator) : ""
    }
    join(dirs, paths[-1])
  end
end

class Suikyo
  attr_reader :table
  def initialize (table = nil)
    if table then
      @table = table
    else
      @table = SuikyoTable2.new
    end
  end

  def convert (string, table = @table)
    string = string.dup
    orig_table = table
    conversion = ""
    last_result = nil
    last_string = nil

    until string.empty? do
      head = string.slice!(0,1)
      node = table.word(head)

      if node.result then
	last_result = node.result
	last_string = (node.cont or "") + string.dup
      elsif last_result == nil then
	last_result = head.dup
	last_string = string.dup
      end

      if node.subtable && (not string.empty?) then
	table = node.subtable
      else
	conversion = conversion + last_result
	string     = last_string
	table       = orig_table
	last_result = nil
	last_string = nil
      end
    end

    return conversion
  end

  def expand (string, table = @table)
    string = string.dup
    orig_table = table
    conversion = ""
    last_result = nil
    last_string = nil

    until string.empty? do
      head = string.slice!(0,1)
      node = table.word(head)

      if node.result then
	last_result = node.result
	last_string = (node.cont or "") + string.dup
      elsif last_result == nil then
	last_result = head.dup
	last_string = string.dup
      end

      if node.subtable then
	table = node.subtable
      else
	conversion = conversion + last_result
	string     = last_string
	table       = orig_table
	last_result = nil
	last_string = nil
      end
    end

    if table != orig_table then
      suffix_list = expand_table(table).push(node.result).compact.uniq
      conversions = suffix_list.map {|suffix|
	conversion + suffix
      }
    else
      conversions = [conversion]
    end
    if last_result != nil then
      conversion += last_result + last_string
    end
    return [conversion, conversions]
  end

  def valid? (string, table = @table)
    # Check a validness of string conversion.
    #   valid: "ringo" -> "$B$j$s$4(B"
    # invalid: "apple" -> "$B$"$C(Bpl$B$((B"
    (conversion, conversions) = expand(string, table)

    # Checking "appl -> $B$"$C(Bpl" (invaild)
    if conversions.length == 1 and conversion !~ /^[a-zA-Z]*[^a-zA-Z]+$/ then
      return false
    end

    conversions.each {|word|
      if word =~ /^[^a-zA-Z]+[a-zA-Z]*$/ then
	return true
      end
    }
    return false
  end

  private
  def expand_table (table)
    return [] unless table

    results = []
    table.allresults_uniq.each {|result, cont|
      if cont then
	if @table.word(cont).subtable then
	  @table.word(cont).subtable.allresults_uniq.each {|subresult, subcont|
	    results.push(result + subresult)
	  }
	else
	  results.push(result + cont)
	end
      else
	results.push(result)
      end
    }
    return results.uniq
  end
end

class SuikyoTable
  attr_reader :table_files

  def initialize
    @default_node = SuikyoNode.new
    @word = Hash.new(@default_node)
    @table_files = []
  end

  def set (string, result, cont = nil, unescape = true)
    if unescape then
      string = unescape(string)
      result = unescape(result)
      cont   = (cont and unescape(cont))
    end

    head = string[0..0]
    rest = string[1..-1]
    @word[head] = SuikyoNode.new if @word[head] == @default_node

    if rest == "" then
      @word[head].result = result
      @word[head].cont   = cont
    else
      @word[head].subtable = self.class.new unless @word[head].subtable
      @word[head].subtable.set(rest, result, cont, false)
    end
  end

  def loadfile (filename, tablepath = nil)
    filepath = SuikyoTable::loadpath(filename, tablepath)
    if FileTest::exist?(filepath) then
      @table_files.push(filepath)
    else
      $stderr.puts "Suikyo.rb: conv-table '#{filepath}' is not found."
      return false
    end

    file = open(filepath, "r")
    comment_flag = false
    file.readlines.each{|line|
      line = line.toeuc.chomp
      if line =~ /^\/\*/ then
	comment_flag = true
      end
      unless line =~ /^\#|^\s*$/ or comment_flag then
	(string, result, cont) = line.sub(/^ /, "").split
	self.set(string, result, cont)
      end
      if line =~ /\*\// then
	comment_flag = false
      end
    }
    file.close
    return true
  end

  def SuikyoTable::loadpath (filename, tablepath = nil)
    if filename =~ /^\// then
      return filename
    else
      prefix = (tablepath or ENV['SUIKYO_TABLE_PATH'] or SUIKYO_TABLE_PATH)
      return File::join2(prefix, filename) 
    end
  end

  def word (char)
    return @word[char]
  end

  def allword
    return @word
  end

  def allresults
    # c => [$B$A(B, $B$A$c(B, $B$A$e(B, $B$A$g(B]
    results = []
    allword.each {|char, table|
      if table.result then
	results.push([table.result, table.cont])
      end
      if table.subtable then
	results += table.subtable.allresults
      end
    }
    return results.uniq
  end

  def allresults_uniq
    # c => [$B$A(B]
    results = allresults.sort {|pair1, pair2|
      pair1[0] <=> pair2[0]
    }
    (base_result, base_cont) = results[0]
    uniq_results = [results[0]]

    results.each {|result, cont|
      unless result.index(base_result) == 0 and cont == base_cont then
	uniq_results.push([result, cont])
	base_result = result
	base_cont   = cont
      end
    }
    return uniq_results
  end


  private
  def unescape (string)
    unescaped_string = ""
    while (index = string.index('\\')) do
      next_char = string[index + 1,1]
      case next_char
      when "x" then
	unescaped_string += string[0,index] + string[index + 2,2].hex.chr
	string = (string[index + 4..-1] or "")
      when "0" then
	unescaped_string += string[0,index]
	string = (string[index + 2..-1] or "")
      else
	unescaped_string += string[0,index] + next_char
	string = (string[index + 2..-1] or "")
      end
    end
    return unescaped_string + string
  end

  private
  class SuikyoNode
    attr_accessor :subtable, :cont, :result
    def initialize (result = nil, cont = nil, subtable = nil)
      @result   = result
      @cont     = cont
      @subtable = subtable
    end
  end
end

class SuikyoTable2 < SuikyoTable
  def word (char)
    if @word[char] != @default_node then
      return @word[char]
    else
      return @word[char.swapcase]
    end
  end
end

